/*

*************************************************************************

ArmageTron -- Just another Tron Lightcycle Game in 3D.
Copyright (C) 2000  Manuel Moos (manuel@moosnet.de)

**************************************************************************

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
  
***************************************************************************

*/

#define HUD_MAP

#ifdef HUD_MAP
#define DONTDOIT
#include "eAdvWall.h"
#include "ePlayer.h"
#include "eRectangle.h"
#include "gWall.h"
#include "gCycle.h"
#include "tConfiguration.h" 
#endif
// I don't know if these includes are needed, I'm brute-forcing it :) - Dave
#include "gMenus.h"
#include "ePlayer.h"
#include "rScreen.h"
#include "nConfig.h"
#include "rConsole.h"
#include "tToDo.h"
#include "rGL.h"
#include "eTimer.h"
#include "eFloor.h"
#include "rRender.h"
#include "rModel.h"
#include "gGame.h"
#include "tRecorder.h"

#include <rRender.h>
#include <math.h>
#include "gCycle.h"
#include <time.h>

#include "gHud.h"

#ifdef HUD_MAP
extern tList<gNetPlayerWall> gridded_sg_netPlayerWalls;

static int simplemapmode = 1;
static tConfItem<int> simplemapmode_con("SIMPLE_MAP", simplemapmode);
#endif

REAL subby_SpeedGaugeSize=.150, subby_SpeedGaugeLocX=-0.165, subby_SpeedGaugeLocY=-0.9;
REAL subby_BrakeGaugeSize=.150, subby_BrakeGaugeLocX=0.25, subby_BrakeGaugeLocY=-0.9;
REAL subby_RubberGaugeSize=.150, subby_RubberGaugeLocX=-0.55, subby_RubberGaugeLocY=-0.9;
bool subby_SpeedGaugeBar=true, subby_BrakeGaugeBar=true, subby_RubberGaugeBar=true;
bool subby_SpeedGaugeHighIsBetter=false, subby_BrakeGaugeHighIsBetter=true, subby_RubberGaugeHighIsBetter=false;
bool subby_SpeedGaugeReverse=false, subby_BrakeGaugeReverse=false, subby_RubberGaugeReverse=false;
bool subby_ShowHUD=true, subby_ShowSpeedFastest=true, subby_ShowScore=true, subby_ShowAliveEnemies=true, subby_ShowPing=true, subby_ShowSpeedMeter=true, subby_ShowBrakeMeter=true, subby_ShowRubberMeter=true;
bool showTime=false;
bool show24hour=false;
REAL subby_ScoreLocX=-0.95, subby_ScoreLocY=-0.80, subby_ScoreSize =.13;
REAL subby_FastestLocX=-0.68, subby_FastestLocY=-0.95, subby_FastestSize =.12;
REAL subby_AliveEnemiesLocX=-0.16, subby_AliveEnemiesLocY=-0.95, subby_AliveEnemiesSize =.13;
REAL subby_PingLocX = 0.30, subby_PingLocY = -0.95, subby_PingSize = 0.13;

class gHudValue {
public:
    gHudValue(){};
    virtual REAL GetReal(gCycle *h){return 0;};
    virtual tString GetString(){return (tString)("");};
};
class gHudValueCycleInfo : public gHudValue {
    tString m_ValueName;
public:
    REAL GetReal(gCycle *h);
    gHudValueCycleInfo(tString name){ m_ValueName = name; };
};
REAL gHudValueCycleInfo::GetReal(gCycle *h){
    if(m_ValueName=="CurrentRubber")
	return(h->GetRubber());
    if(m_ValueName=="CurrentSpeed")
	return(h->Speed());
    if(m_ValueName=="CurrentBrakes")
	return(h->GetBrakingReservoir());
    std::cout << "Undefined value: " << m_ValueName << "\n";
}

class gHudGauge : public gHudWidget {
public:
    gHudGauge();
    gHudGauge(const char* title, float size, float min, float max, float x, float y, float z=0.0);
    ~gHudGauge() { if(m_ValueSouce) {delete m_ValueSouce; m_ValueSouce=0;} };
    void Render(gCycle *h);
    void SetMin(float min) { m_Min = min; };
    void SetMax(float max) { m_Max = max; };
    void SetValue(float value) { m_Value = value; };
    void SetValueSourceCycleInfo(tString name) { m_ValueSouce = new gHudValueCycleInfo(name); };
private:
    float m_Value;
    gHudValue *m_ValueSouce;
    tColoredString m_Title;
    float m_Min;
    float m_Max;
    float m_X;
    float m_Y;
    float m_Z;
    float m_Size;
    bool m_Usebar;
    bool m_Reverse;
    REAL r;
    REAL g;
    REAL b;
    bool m_DisplayValue;
    gCycle *m_Cycle;
};

gHudGauge::gHudGauge(const char* title, float size, float min, float max, float x, float y, float z)
        : m_Size(size), m_X(x), m_Y(y), m_Z(z), m_Min(min), m_Max(max),
        m_Usebar(true), m_Reverse(false),
        r(0.5), g(0.5), b(0.5), m_DisplayValue(true), m_ValueSouce(0) {
    // Set up the title
    tString tmpTitle(title);
    tOutput titleLocal(tmpTitle);
        
    m_Title << "0xff3333" << titleLocal;
}

void gHudGauge::Render(gCycle *h) {
    sr_ResetRenderState(0); //needs this because rFont has bugs i presume.. Ie I have problems as soon as rTextFirld is used
    // z-man: actually, it is needed because per-frame-tasks get called without rendering context, so it has to be set.

    if (m_ValueSouce)
	m_Value = m_ValueSouce->GetReal(h);

    float x, y;
    char string[50];
    bool displayvalue = m_DisplayValue;
    m_Value > m_Max ? m_Value = m_Max:1;
    m_Value < 0 ? m_Value=0:1;
    if(m_Usebar) {
        int factor=( m_Reverse?-1:1);
        x= factor * (m_Value/m_Max*2. - 1.);
        //render the box
        BeginQuads();
        Color(r,g, b,0.1);
        Vertex(m_Size + m_X, m_Size * 0.3 + m_Y, 0.0);
        Vertex(m_Size + m_X, m_Y , 0.0);
        Vertex(m_Size * x + m_X, m_Y, 0.0);
        Vertex(m_Size * x + m_X, m_Size * 0.3 + m_Y, 0.0);
        RenderEnd();
        //fill
        BeginQuads();
        //if(highisbetter)
        //    Color((x>0?1.-x:1.),(x<0.?x+1.:1.),0.,.5);
        //else
            Color((x<0.?x+1.:1.),(x>0?1.-x:1.),0.,.5);
        Vertex(-m_Size+m_X, m_Size*0.3+m_Y,0);
        Vertex(-m_Size+m_X, m_Y,0);
        Vertex(m_Size*x+m_X,m_Y,0);
        Vertex(m_Size*x+m_X,m_Size*0.3+m_Y,0);
        RenderEnd();
        //render the line
        BeginLines();
        Color(r,g, b);
        Vertex(m_Size*x+m_X,m_Y,0);
        Vertex(m_Size*x+m_X, m_Size*0.3+m_Y,0);
        RenderEnd();
    } else {
        x= (m_Reverse?-1:1) * cos(M_PI*m_Value/m_Max);
        y= sin(M_PI*m_Value/m_Max);
        //if(y < m_Size*.24) displayvalue = false; // dont display text when at minimum
    
        /* Draws an ugly background on the gauge
            BeginQuads();
            Color(1.,1.,1.,.8);
            Vertex(m_X-m_Size-.04,m_Y-.04,0);
            Vertex(m_X-m_Size-.04,m_Y+m_Size+.04,0);
            Vertex(m_X+m_Size+.04,m_Y+m_Size+.04,0);
            Vertex(m_X+m_Size+.04,m_Y-.04,0);
            
            Color(.1,.1,.1,.8);
            Vertex(m_X-m_Size-.02,m_Y-.02,0);
            Vertex(m_X-m_Size-.02,m_Y+m_Size+.02,0);
            Vertex(m_X+m_Size+.02,m_Y+m_Size+.02,0);
            Vertex(m_X+m_Size+.02,m_Y-.02,0);
    
            RenderEnd();*/
    
        BeginLines();
        Color(r,g, b);
        Vertex(-.1*x*m_Size+m_X,.1*y*m_Size+m_Y,0);
        Vertex(-x*m_Size+m_X,y*m_Size+m_Y,0);
        RenderEnd();
    }


    rTextField min_t(-m_Size-(0.1*m_Size)+m_X,m_Y,.12*m_Size,.24*m_Size);
    rTextField max_t(+m_Size+(0.1*m_Size)+m_X,m_Y,.12*m_Size,.24*m_Size);
    min_t << "0xcccccc" << (m_Reverse?m_Max:0);
    max_t << "0xcccccc" << (m_Reverse?0:m_Max);

    if(displayvalue){
        if(m_Usebar){
            int length = sprintf(string,"%.1f",m_Value);
            rTextField speed_t(m_X-((.15*m_Size*(length-1.5))/2.0),m_Y+m_Size*.28,.12*m_Size,.24*m_Size);
            speed_t << "0xccffff" << (string);
        }else{
            rTextField speed_t(-x*1.45*m_Size+m_X,y*1.35*m_Size+m_Y,.1*m_Size,.2*m_Size);
            sprintf(string,"%.1f",m_Value);
            speed_t << "0xccffff" << (string);
        }
    }
    int length = m_Title.Len();
    if ( length > 0 )
    {
        //centre  -1.0 for null char and -.5 for half a char width = -1.5
        rTextField titletext(m_X-((.15*m_Size*(length-1.5))/2.0),m_Y,.12*m_Size,.24*m_Size); 
        titletext << m_Title;
    }
#ifndef DEDICATED
    if (!sr_glOut)
        return;

    glColor3f(1,1,1);
#endif
}

gHud* gHud::_instance = 0;

gHud::gHud() {
#ifndef DEDICATED
    // Load the xml descriptor file for the hud layout
    // TODO
    m_Player = NULL;

    gHudGauge *gauge;
    // This initialisation is to be done by the XML parser, so this is temporary
    if(subby_ShowRubberMeter){
	gauge = new gHudGauge("Rubber", subby_RubberGaugeSize, 0, 10, subby_RubberGaugeLocX, subby_RubberGaugeLocY);
	gauge->SetMin(0.0);
	gauge->SetValueSourceCycleInfo((tString)"CurrentRubber");
	gauge->SetMax(sg_rubberCycle);
	gHudWidgetContainer container(gauge);
	Widgets.push_back(gauge);
    }
    if(subby_ShowSpeedMeter){
	gauge = new gHudGauge("Speed", subby_SpeedGaugeSize, 0, 10, subby_SpeedGaugeLocX, subby_SpeedGaugeLocY);
	gauge->SetMin(0.0);
	gauge->SetValueSourceCycleInfo((tString)"CurrentSpeed");
	gauge->SetMax(60.0);
	gHudWidgetContainer container(gauge);
	Widgets.push_back(gauge);
    }
    if(subby_ShowBrakeMeter){
	gauge = new gHudGauge("Brake", subby_BrakeGaugeSize, 0, 10, subby_BrakeGaugeLocX, subby_BrakeGaugeLocY);
	gauge->SetMin(0.0);
	gauge->SetValueSourceCycleInfo((tString)"CurrentBrakes");
	gauge->SetMax(1.0);
	gHudWidgetContainer container(gauge);
	Widgets.push_back(gauge);
    }
#endif
}

gHud* gHud::GetHud() {
    if(!_instance)
        _instance = new gHud();

    return _instance;
}

void gHud::RenderEnemies() {
#ifndef DEDICATED
    int alivepeople=0;
    int alivemates=0;
    unsigned short int max = se_PlayerNetIDs.Len();
    ePlayerNetID *me = m_Player->netPlayer;
        
    if (me){
        for(unsigned short int i=0;i<max;i++){
            ePlayerNetID *p=se_PlayerNetIDs(i);
            if (p->Object() && p->Object()->Alive()){
                if (p->CurrentTeam() != me->CurrentTeam()){
                    alivepeople++;
                }else{
                    alivemates++;
                }
            }
        }
    }
    
    if(subby_ShowAliveEnemies){
        tString message;
        message << tOutput("$hud_enemies_alive");
        message << " " << alivepeople << " " << tOutput("$hud_teammates_alive");
        message << " " << alivemates;
        int length = message.Len();
        float size = subby_AliveEnemiesSize;
        rTextField enemies_alive(subby_AliveEnemiesLocX-((.15*size*(length-1.5))/2.0),subby_AliveEnemiesLocY,.15*size,.3*size);
        enemies_alive << "0xfefefe" << message;
    }
#endif
}

void gHud::RenderPing() {
#ifndef DEDICATED
    ePlayerNetID *me = m_Player->netPlayer;
    float myping;
    myping = me->ping;
    if(subby_ShowPing){
        tString message;
        message << tOutput("$hud_ping");
        message << " " << int(myping * 1000) << " ms" ;
        int length = message.Len();
        float size = subby_PingSize;
        rTextField ping(subby_PingLocX-((.15*size*(length-1.5))/2.0),subby_PingLocY,.15*size,.3*size);
        ping << "0xfefefe" << message;
    }
#endif
}

void gHud::RenderScore() {
#ifndef DEDICATED
static REAL myscore, topscore;
    topscore =0;
    ePlayerNetID *me = m_Player->netPlayer;
    ePlayerNetID *p;
        
    if ( me )
        myscore = me->TotalScore();
    
    for(int i =0 ; i< se_PlayerNetIDs.Len(); i++){
        p = se_PlayerNetIDs[i];
        p->TotalScore()>topscore?topscore=p->TotalScore():1;
    }
    myscore=p->TotalScore();
    if(subby_ShowScore){
        tString colour;
        if(myscore==topscore){
            colour = "0xff9d50";
        }else if (myscore > topscore){
            colour = "0x11ff11";
        }else{
            colour = "0x11ffff";
        }

        float size = subby_ScoreSize;
        rTextField score(subby_ScoreLocX,subby_ScoreLocY,.15*size,.3*size);
        score <<               tOutput("$hud_score");
        score << "\n0xefefef" << tOutput("$hud_score_player");
        score << "  " << tOutput("$hud_score_top");
        score << "\n" << colour << myscore << "     0xffff00" << topscore ;
    }
#endif
}

void gHud::RenderFPS() {
#ifndef DEDICATED
if (!(se_mainGameTimer &&
            se_mainGameTimer->speed > .9 &&
            se_mainGameTimer->speed < 1.1 &&
            se_mainGameTimer->IsSynced() ) )
        return;

    static int fps       = 60;
    static REAL lastTime = 0;

    REAL newtime = tSysTimeFloat();
    REAL ts      = newtime - lastTime;

    float size =.15;
    rTextField c2(.7,.85,.15*size, .3*size);

    if ( tRecorder::IsRunning() )
    {
        std::stringstream time;
        time.precision( 5 );
        time << newtime;
        {
            c2 << "0xffffffT:" << time.str() << "\n";
        }
    }

    int newfps   = static_cast<int>(se_AverageFPS());
    if (fabs((newfps-fps)*ts)>4)
    {
        fps      = newfps;
        lastTime = newtime;
    }

    if(sr_FPSOut){
        c2 << "0xffffffFPS: " <<fps;
    }
    // Show the time
    if(showTime) {
        static int lastTime=0;
        static char theTime[13];
        float size =.15;
        struct tm* thisTime;
        time_t rawtime;

        time ( &rawtime );
        thisTime = localtime ( &rawtime );

        if(thisTime->tm_min != lastTime) {
            char h[3];
            char m[3];
            char ampm[3] = " ";

            lastTime = thisTime->tm_min;

            if(thisTime->tm_min < 10) {
                sprintf(m, "0%d", thisTime->tm_min);
            } else {
                sprintf(m, "%d", thisTime->tm_min);
            }

            if(show24hour) {
                sprintf(h, "%d", thisTime->tm_hour);
            } else {
                int newhour;

                if(thisTime->tm_hour > 12) {
                    newhour = thisTime->tm_hour-12;
                    sprintf(ampm,"%s","PM");
                } else {
                    newhour = thisTime->tm_hour;
                    if(newhour == 0) newhour = 12;
                    sprintf(ampm,"%s","AM");
                }
                sprintf(h, "%d", newhour);
            }

            sprintf(theTime, "%s:%s %s",h,m,ampm);
        }

        // float timesize = subby_ScoreSize; // -((.15*timesize*(10-1.5))/2.0)
        rTextField the_time(.7,.9,.15*size, .3*size);

        the_time << "0xffffff" << theTime;
    }
#endif
}

void gHud::RenderFastest() {
#ifndef DEDICATED
// static int max=0,lastmax=0;
    // static int imax=-1,lastimax=0;
    ePlayerNetID *me = m_Player->netPlayer;
    const eGameObject *ret = NULL;
    static REAL max=0;
    static REAL imax=-1;
    char fasteststring[50];
    // REAL distance;
    static float maxmeterspeed= 50;
    static tString name;
    static tString ultiname;
    static bool  wrotefastest = false;
    static bool wrote10 =false;
    static bool belowzero=false;
    static REAL ultimax=0;
    static REAL timelast;
    // tString realname = ePlayer::PlayerConfig(0)->Name();

    // const tString& myname = me->name;
    for(int i =0 ; i< se_PlayerNetIDs.Len(); i++){
        ePlayerNetID *p = se_PlayerNetIDs[i];

        if(gCycle *h = (gCycle*)(p->Object())){

            // change player so we always see the gauges of the cycle that is watched
            if ( h == ret )
                me = p;

            //   distance = h->distance;
            //   c << p-> name << " " << int((h ->Speed())) << " ";
            if (h->Speed()>max){
                max =  (float) h->Speed();  // changed to float for more accuracy in reporting top speed
                name = p->GetName();
                imax = i;

            }
            if( h->Speed()>ultimax){
                ultimax = (float) h->Speed();
                ultiname = p->GetName();
            }
        }
    }
    
    if(subby_ShowSpeedFastest)
    {

        float size= subby_FastestSize;
        tColoredString message,messageColor;
        messageColor << "0xbf9d50";

        sprintf(fasteststring,"%.1f",max);
        message << "  " << tOutput("$hud_fastest");
        message << " " << name << " " << fasteststring;
        message.RemoveHex(); //cheers tank;
        int length = message.Len();

        rTextField speed_fastest(subby_FastestLocX-((.15*size*(length-1.5))/2.0),subby_FastestLocY,.15*size,.3*size);
        /*   rTextField speed_fastest(.7-((.15*size*(length-1.5))/2.0),.65,.15*size,.3*size); */

        speed_fastest << messageColor << message;
    }
#endif
}

void gHud::RenderOld() {
#ifndef DEDICATED
    if( m_Player == NULL ) return;
        
    sr_ResetRenderState(true);
    RenderFPS();

    ePlayerNetID *me = m_Player->netPlayer;
            
    if(me != NULL) {
        if (se_mainGameTimer &&
                se_mainGameTimer->speed > .9 &&
                se_mainGameTimer->speed < 1.1 &&
                se_mainGameTimer->IsSynced() )
        {
            sr_ResetRenderState(false);
            Color(1,1,1);
            //rTextField c(-.9,-.85);
            //rTextField t(.6,.98);
    
            // t << "0xffffffGame Time:" << int (se_GameTime())<< " s";
            if(subby_ShowHUD){
    
                if(m_Player->cam) {
                    //ret = m_Player->cam->Center();
    
                    static bool firsttime =true;
    
                    /*if(se_GameTime()<0||!firsttime){
                        firsttime=false;
                        if((se_GameTime()<0)){
                            belowzero=true;
                            if(se_GameTime()<-1){
                                tOutput message;
                                //  message<< "Last Game Fastest: " << name << " with a top speed of " << max << "\n";
                                //  con << "testcon";
                                //  sn_CenterMessage(message);
                            }else{
                                timelast = se_GameTime();
                                max=0;
                                wrotefastest = false;
                                wrote10 =false;
                            }
                        }else if(se_GameTime()>0&&belowzero){
                            belowzero=false;
                        }
                    }*/
    
                    if (me!=NULL){
                        gCycle *h = dynamic_cast<gCycle *>(me->Object());
                        if (h && ( !m_Player->netPlayer || !m_Player->netPlayer->IsChatting()) && se_GameTime()>-2){
                            //h->Speed()>maxmeterspeed?maxmeterspeed+=10:1;
                            
    
			    for(int i=0; i<Widgets.size(); i++)
			    {
				Widgets[i].GetWidget()->Render(h);
			    }
                            //  bool displayfastest = true;// put into global, set via menusytem... subby to do.make sr_DISPLAYFASTESTout
    
                            RenderScore();
                            RenderPing();
                            RenderEnemies();
                        }
                    }
                }
            }
        }
    }
#endif
}


#ifndef DEDICATED
#if 0
void GLmeter_subby(float value,float max, float locx, float locy, float size, const char * t,bool displayvalue = true, bool reverse = false, REAL r=.5, REAL g=.5, REAL b=1, bool highisbetter=false, bool usebar=true)
{
    tString title(t);
    tOutput titleLocal(title);
    title = titleLocal;

    sr_ResetRenderState(0); //needs this because rFont has bugs i presume.. Ie I have problems as soon as rTextFirld is used
    // z-man: actually, it is needed because per-frame-tasks get called without rendering context, so it has to be set.
    float x, y;
    char string[50];
    value>max?value=max:1;
    value<0?value=0:1;
    if(usebar) {
	int factor=(reverse?-1:1);
	x= factor * (value/max*2. - 1.);
	//render the box
	BeginQuads();
	Color(r,g, b,0.1);
	Vertex(size+locx, size*0.3+locy,0);
	Vertex(size+locx, locy,0);
	Vertex(size*x+locx,locy,0);
	Vertex(size*x+locx,size*0.3+locy,0);
	RenderEnd();
	//fill
	BeginQuads();
	if(highisbetter)
	    Color((x>0?1.-x:1.),(x<0.?x+1.:1.),0.,.5);
	else
	    Color((x<0.?x+1.:1.),(x>0?1.-x:1.),0.,.5);
	Vertex(-size+locx, size*0.3+locy,0);
	Vertex(-size+locx, locy,0);
	Vertex(size*x+locx,locy,0);
	Vertex(size*x+locx,size*0.3+locy,0);
	RenderEnd();
	//render the line
	BeginLines();
	Color(r,g, b);
	Vertex(size*x+locx,locy,0);
	Vertex(size*x+locx, size*0.3+locy,0);
	RenderEnd();
    }else{
	x= (reverse?-1:1) * cos(M_PI*value/max);
	y= sin(M_PI*value/max);
	if(y<size*.24) displayvalue = false; // dont display text when at minimum

	/* Draws an ugly background on the gauge
	    BeginQuads();
	    Color(1.,1.,1.,.8);
	    Vertex(locx-size-.04,locy-.04,0);
	    Vertex(locx-size-.04,locy+size+.04,0);
	    Vertex(locx+size+.04,locy+size+.04,0);
	    Vertex(locx+size+.04,locy-.04,0);
	    
	    Color(.1,.1,.1,.8);
	    Vertex(locx-size-.02,locy-.02,0);
	    Vertex(locx-size-.02,locy+size+.02,0);
	    Vertex(locx+size+.02,locy+size+.02,0);
	    Vertex(locx+size+.02,locy-.02,0);

	    RenderEnd();*/

	BeginLines();
	Color(r,g, b);
	Vertex(-.1*x*size+locx,.1*y*size+locy,0);
	Vertex(-x*size+locx,y*size+locy,0);
	RenderEnd();
    }


    rTextField min_t(-size-(0.1*size)+locx,locy,.12*size,.24*size);
    rTextField max_t(+size+(0.1*size)+locx,locy,.12*size,.24*size);
    min_t << "0xcccccc" << (reverse?max:0);
    max_t << "0xcccccc" << (reverse?0:max);

    if(displayvalue){
	if(usebar){
	    int length = sprintf(string,"%.1f",value);
	    rTextField speed_t(locx-((.15*size*(length-1.5))/2.0),locy+size*.28,.12*size,.24*size);
	    speed_t << "0xccffff" << (string);
	}else{
	    rTextField speed_t(-x*1.45*size+locx,y*1.35*size+locy,.1*size,.2*size);
	    sprintf(string,"%.1f",value);
	    speed_t << "0xccffff" << (string);
	}
    }
    int length = title.Len();
    if ( length > 0 )
    {
        rTextField titletext(locx-((.15*size*(length-1.5))/2.0),locy,.12*size,.24*size); //centre  -1.0 for null char and -.5 for half a char width = -1.5
        titletext << "0xff3333" << titleLocal;
    }

#ifndef DEDICATED
    if (!sr_glOut)
        return;

    glColor3f(1,1,1);
#endif

}

static int alivepeople, alivemates, thetopscore, hudfpscount;

static void tank_display_hud( ePlayerNetID* me ){
    static int fps       = 60;
    static REAL lastTime = 0;

    if (se_mainGameTimer &&
            se_mainGameTimer->speed > .9 &&
            se_mainGameTimer->speed < 1.1 &&
            se_mainGameTimer->IsSynced() )
    {
        REAL newtime = tSysTimeFloat();
        REAL ts      = newtime - lastTime;
        int newfps   = static_cast<int>(se_AverageFPS());
        if (fabs((newfps-fps)*ts)>4)
        {
            fps      = newfps;
            lastTime = newtime;
        }

        sr_ResetRenderState(false);
        Color(1,1,1);
        rTextField c(-.9,-.6);

        // signed short int playerID = -1;
        //	  unsigned short int alivepeople2 = 0;
        unsigned short int max = se_PlayerNetIDs.Len();
        //	  signed short int scores[16];
        //	  signed short int teamID = -1;

        bool dohudcrap = false;
        if (hudfpscount >= fps) {
            hudfpscount = 0;
            dohudcrap = true;
        }
        else { hudfpscount++; }


        //Calculation, enemies (people not on me's team) alive
        if (dohudcrap) {
            alivepeople=0;
            alivemates=0;
            if (me){
                for(unsigned short int i=0;i<max;i++){
                    ePlayerNetID *p=se_PlayerNetIDs(i);
                    if (p->Object() && p->Object()->Alive()){
                       if (p->CurrentTeam() != me->CurrentTeam()){
                           alivepeople++;
                       }else{
                           alivemates++;
                       }
                    }
                }
            }
        }

        /*	  int max = teams.Len();
        	  for(int i=0;i<max;i++){
        	    eTeam *t = teams(i);
        	  }*/

        if ( me ){
            //Calculation, top player score
            if (dohudcrap) {
                //	    unsigned short int maxscore = 0;
                thetopscore = 0;
                for(unsigned short int i=0;i<max;i++){
                    ePlayerNetID* p2=se_PlayerNetIDs(i);
                    //	  if (p2->TotalScore() > p->TotalScore() || p2->TotalScore() > thetopscore){
                    if (p2->TotalScore() > thetopscore){
                        thetopscore = p2->TotalScore();
                    }
                }
            }

            eCoord pcpos; // = 0 implied
            if (me->Object() && me->Object()->Alive()) {
                pcpos = me->Object()->Position(); }

            tString line1, line2, line3;

            /*    line1 << "HUD: " << p->name;
              line1.SetPos(22, true);
                    line1 << "FPS: " << fps;
              line1.SetPos(34, true);
              line1 << "ENEMIES: " << alivepeople << "\n";

                 line2 << "SCORE: " << p->TotalScore() << ", " << thetopscore;
              line2.SetPos(22, true);
                 line2 << "PING: " << int(p->ping*1000) << "\n";

              line3 << "LOCATION: " << int(pcpos.x);
              line3 << ", " << int(pcpos.y);
              line3.SetPos(22, true);
                // line3 << "RUBBER: " << int(p->rubberstatus*100);


             c <<line1 << line2; */


        }
    }
}

static void display_hud_subby( ePlayer* player ){
    if ( !player )
        return;
    ePlayerNetID *me = player->netPlayer;

    if (se_mainGameTimer &&
            se_mainGameTimer->speed > .9 &&
            se_mainGameTimer->speed < 1.1 &&
            se_mainGameTimer->IsSynced() )
    {
        sr_ResetRenderState(false);
        Color(1,1,1);
        rTextField c(-.9,-.85);
        rTextField t(.6,.98);

        // t << "0xffffffGame Time:" << int (se_GameTime())<< " s";
        if(subby_ShowHUD){
            // static int max=0,lastmax=0;
            // static int imax=-1,lastimax=0;
            static REAL max=0;
            static REAL imax=-1;
            char fasteststring[50];
            // REAL distance;
            static float maxmeterspeed= 50;
            float myping;
            static tString name;
            static tString ultiname;
            static bool  wrotefastest = false;
            static bool wrote10 =false;
            static bool belowzero=false;
            static REAL ultimax=0;
            static REAL timelast;
            static REAL myscore, topscore;
            // tString realname = ePlayer::PlayerConfig(0)->Name();

            // const tString& myname = me->name;

            if ( me )
                myscore = me->TotalScore();

            const eGameObject *ret = NULL;
            if(player->cam){
                ret = player->cam->Center();

                static bool firsttime =true;

                if(se_GameTime()<0||!firsttime){
                    firsttime=false;
                    if((se_GameTime()<0)){
                        belowzero=true;
                        if(se_GameTime()<-1){
                            tOutput message;
                            //  message<< "Last Game Fastest: " << name << " with a top speed of " << max << "\n";
                            //  con << "testcon";
                            //  sn_CenterMessage(message);
                        }else{
                            timelast = se_GameTime();
                            max=0;
                            wrotefastest = false;
                            wrote10 =false;
                        }
                    }else if(se_GameTime()>0&&belowzero){
                        belowzero=false;
                    }
                }

                topscore =0;
                for(int i =0 ; i< se_PlayerNetIDs.Len(); i++){
                    ePlayerNetID *p = se_PlayerNetIDs[i];
                    p->TotalScore()>topscore?topscore=p->TotalScore():1;

                    if(gCycle *h = (gCycle*)(p->Object())){

                        // change player so we always see the gauges of the cycle that is watched
                        if ( h == ret )
                            me = p;

                        //   distance = h->distance;
                        //   c << p-> name << " " << int((h ->Speed())) << " ";
                        if (h->Speed()>max){
                            max =  (float) h->Speed();  // changed to float for more accuracy in reporting top speed
                            name = p->GetName();
                            imax = i;

                        }
                        if( h->Speed()>ultimax){
                            ultimax = (float) h->Speed();
                            ultiname = p->GetName();
                        }
                    }
                }

                if (me!=NULL){
                    gCycle *h = dynamic_cast<gCycle *>(me->Object());
                    if (h && ( !player->netPlayer || !player->netPlayer->IsChatting()) && se_GameTime()>-2){
                        h->Speed()>maxmeterspeed?maxmeterspeed+=10:1;
                        // myscore=p->TotalScore();
                        myping = me->ping;

			//just for me, can be removed
			//void GLmeter_subby(float value,float max, float locx, float locy, float size, const char * t,bool displayvalue = true, bool reverse = false, REAL r=.5, REAL g=.5, REAL b=1, bool highisbetter=false, bool usebar=true)
                        if(subby_ShowSpeedMeter)
                            GLmeter_subby(h->Speed(),maxmeterspeed,subby_SpeedGaugeLocX,subby_SpeedGaugeLocY,subby_SpeedGaugeSize,"Speed",true, false, .5, .5, 1, subby_SpeedGaugeHighIsBetter, subby_SpeedGaugeBar);  // easy to use configurable meters
                        if(subby_ShowRubberMeter)
                        {
                            GLmeter_subby(h->GetRubber(),sg_rubberCycle,subby_RubberGaugeLocX,subby_RubberGaugeLocY,subby_RubberGaugeSize," Rubber Used", true, false, .5, .5, 1., subby_RubberGaugeHighIsBetter, subby_RubberGaugeBar);
                            if ( gCycle::RubberMalusActive() )
                                GLmeter_subby(100/(1+h->GetRubberMalus()),100,subby_RubberGaugeLocX,subby_RubberGaugeLocY,subby_RubberGaugeSize,"",true, true);
                        }
                        if(subby_ShowBrakeMeter)
                            GLmeter_subby(h->GetBrakingReservoir(), 1.0,subby_BrakeGaugeLocX,subby_BrakeGaugeLocY,subby_BrakeGaugeSize, " Brakes", true, false, .5, .5, 1., subby_BrakeGaugeHighIsBetter, subby_BrakeGaugeBar);

                        //  bool displayfastest = true;// put into global, set via menusytem... subby to do.make sr_DISPLAYFASTESTout

                        if(subby_ShowSpeedFastest)
                        {

                            float size= subby_FastestSize;
                            tColoredString message,messageColor;
                            messageColor << "0xbf9d50";

                            sprintf(fasteststring,"%.1f",max);
                            message << "  " << tOutput("$hud_fastest");
                            message << " " << name << " " << fasteststring;
                            message.RemoveHex(); //cheers tank;
                            int length = message.Len();

                            rTextField speed_fastest(subby_FastestLocX-((.15*size*(length-1.5))/2.0),subby_FastestLocY,.15*size,.3*size);
                            /*   rTextField speed_fastest(.7-((.15*size*(length-1.5))/2.0),.65,.15*size,.3*size); */

                            speed_fastest << messageColor << message;



                        }

                        if(subby_ShowScore){
                            tString colour;
                            if(myscore==topscore){
                                colour = "0xff9d50";
                            }else if (myscore > topscore){
                                colour = "0x11ff11";
                            }else{
                                colour = "0x11ffff";
                            }

                            float size = subby_ScoreSize;
                            rTextField score(subby_ScoreLocX,subby_ScoreLocY,.15*size,.3*size);
                            score <<               tOutput("$hud_score");
                            score << "\n0xefefef" << tOutput("$hud_score_player");
                            score << "  " << tOutput("$hud_score_top");
                            score << "\n" << colour << myscore << "     0xffff00" << topscore ;
                        }

                        if(subby_ShowAliveEnemies){
                            tString message;
                            message << tOutput("$hud_enemies_alive");
                            message << " " << alivepeople << " " << tOutput("$hud_teammates_alive");
                            message << " " << alivemates;
                            int length = message.Len();
                            float size = subby_AliveEnemiesSize;
                            rTextField enemies_alive(subby_AliveEnemiesLocX-((.15*size*(length-1.5))/2.0),subby_AliveEnemiesLocY,.15*size,.3*size);
                            enemies_alive << "0xfefefe" << message;
                        }

                        if(subby_ShowPing){
                            tString message;
                            message << tOutput("$hud_ping");
                            message << " " << int(myping * 1000) << " ms" ;
                            int length = message.Len();
                            float size = subby_PingSize;
                            rTextField ping(subby_PingLocX-((.15*size*(length-1.5))/2.0),subby_PingLocY,.15*size,.3*size);
                            ping << "0xfefefe" << message;
                        }
                    }
                }
            }
        }

    }
    tank_display_hud( me ); // call tank's version

}

static void display_fps_subby()
{
    if (!(se_mainGameTimer &&
            se_mainGameTimer->speed > .9 &&
            se_mainGameTimer->speed < 1.1 &&
            se_mainGameTimer->IsSynced() ) )
        return;

    static int fps       = 60;
    static REAL lastTime = 0;

    REAL newtime = tSysTimeFloat();
    REAL ts      = newtime - lastTime;

    float size =.15;
    rTextField c2(.7,.85,.15*size, .3*size);

    if ( tRecorder::IsRunning() )
    {
        std::stringstream time;
        time.precision( 5 );
        time << newtime;
        {
            c2 << "0xffffffT:" << time.str() << "\n";
        }
    }

    int newfps   = static_cast<int>(se_AverageFPS());
    if (fabs((newfps-fps)*ts)>4)
    {
        fps      = newfps;
        lastTime = newtime;
    }

    if(sr_FPSOut){
        c2 << "0xffffffFPS: " <<fps;
    }
    // Show the time
    if(showTime) {
        static int lastTime=0;
        static char theTime[13];
        float size =.15;
        struct tm* thisTime;
        time_t rawtime;

        time ( &rawtime );
        thisTime = localtime ( &rawtime );

        if(thisTime->tm_min != lastTime) {
            char h[3];
            char m[3];
            char ampm[3] = " ";

            lastTime = thisTime->tm_min;

            if(thisTime->tm_min < 10) {
                sprintf(m, "0%d", thisTime->tm_min);
            } else {
                sprintf(m, "%d", thisTime->tm_min);
            }

            if(show24hour) {
                sprintf(h, "%d", thisTime->tm_hour);
            } else {
                int newhour;

                if(thisTime->tm_hour > 12) {
                    newhour = thisTime->tm_hour-12;
                    sprintf(ampm,"%s","PM");
                } else {
                    newhour = thisTime->tm_hour;
                    if(newhour == 0) newhour = 12;
                    sprintf(ampm,"%s","AM");
                }
                sprintf(h, "%d", newhour);
            }

            sprintf(theTime, "%s:%s %s",h,m,ampm);
        }

        // float timesize = subby_ScoreSize; // -((.15*timesize*(10-1.5))/2.0)
        rTextField the_time(.7,.9,.15*size, .3*size);

        the_time << "0xffffff" << theTime;
    }

}

static void display_hud_subby_all()
{
    sr_ResetRenderState(true);
    display_fps_subby();

    rViewportConfiguration* viewportConfiguration = rViewportConfiguration::CurrentViewportConfiguration();

    for ( int viewport = viewportConfiguration->num_viewports-1; viewport >= 0; --viewport )
    {
        // get the ID of the player in the viewport
        int playerID = sr_viewportBelongsToPlayer[ viewport ];

        // get the player
        ePlayer* player = ePlayer::PlayerConfig( playerID );

        // select the corrected viewport
        viewportConfiguration->Port( viewport )->CorrectAspectBottom().Select();

        // delegate
        display_hud_subby( player );
    }
}
#endif

static void display_hud_lucifer() {
    gHud* theHud = gHud::GetHud();
    sr_ResetRenderState(true);

    rViewportConfiguration* viewportConfiguration = rViewportConfiguration::CurrentViewportConfiguration();

    for ( int viewport = viewportConfiguration->num_viewports-1; viewport >= 0; --viewport )
    {
        // get the ID of the player in the viewport
        int playerID = sr_viewportBelongsToPlayer[ viewport ];
                
        // get the player
        ePlayer* player = ePlayer::PlayerConfig( playerID );
                
        // select the corrected viewport
        viewportConfiguration->Port( viewport )->CorrectAspectBottom().Select();

        // delegate
        theHud->SetPlayer( player );
        theHud->RenderOld();
    }
}

//static rPerFrameTask dfps(&display_hud_subby_all);
static rPerFrameTask dfps(&display_hud_lucifer);

#ifdef HUD_MAP
static void DrawRimWalls(tList<eWallRim> &list) {
    glColor4f(1, 1, 1, .5);
    glBegin(GL_LINES);
    unsigned i, len=list.Len();
    for(i=0; i<len; i++) {
        eWallRim *wall = list[i];
        const eCoord &begin = wall->EndPoint(0), &end = wall->EndPoint(1);
        glVertex2f(begin.x, begin.y);
        glVertex2f(end.x, end.y);
    }
    glEnd();
}

static void DrawWalls(tList<gNetPlayerWall> &list) {
    unsigned i, len=list.Len();
    double currentTime = se_GameTime();
    double wallsLength = gCycle::WallsLength();
    double wallsStayUpDelay = gCycle::WallsStayUpDelay();
    bool limitedLength = wallsLength > 0;
    glBegin(GL_LINES);
    for(i=0; i<len; i++) {
        gNetPlayerWall *wall = list[i];
        gCycle *cycle = wall->Cycle();
        if(!cycle) continue;
        double alpha = 1;
        if(!cycle->Alive()) {
            alpha -= 2 * (currentTime - cycle->DeathTime() - wallsStayUpDelay);
            if(alpha <= 0) continue;
        }
        glColor4f(cycle->color_.r, cycle->color_.g, cycle->color_.b, alpha);
        double cycleDist = cycle->GetDistance();
        double minDist = limitedLength && cycleDist > wallsLength ? cycleDist - wallsLength : 0;
        const eCoord &begPos = wall->EndPoint(0), &endPos = wall->EndPoint(1);
        tArray<gPlayerWallCoord> &coords = wall->Coords();
        double begDist = wall->BegPos();
        double lenDist = wall->EndPos() - begDist;
        unsigned j, numcoords = coords.Len();
        if(numcoords < 2) continue;
        bool prevDangerous = coords[0].IsDangerous;
        double prevDist = coords[0].Pos;
        if(prevDist < minDist) prevDist = minDist;
        prevDist = (prevDist - begDist) / lenDist;
        for(j=1; j<numcoords; j++) {
            bool curDangerous = coords[j].IsDangerous;
            double curDist = coords[j].Pos;
            if(curDist < minDist) curDist = minDist;
            curDist = (curDist - begDist) / lenDist;
            glVertex2f((1 - prevDist) * begPos.x + prevDist * endPos.x, (1 - prevDist) * begPos.y + prevDist * endPos.y);
            glVertex2f((1 - curDist) * begPos.x + curDist * endPos.x, (1 - curDist) * begPos.y + curDist * endPos.y);
            prevDangerous = curDangerous;
            prevDist = curDist;
        }
    }
    glEnd();
}

static void DrawCycles(tList<ePlayerNetID> &list, double xscale, double yscale) {
    unsigned i, len=list.Len();
    double currentTime = se_GameTime();
    for(i=0; i<len; i++) {
        const gCycle *cycle = dynamic_cast<gCycle *>(list[i]->Object());
        if(cycle) {
            double alpha = 1;
            if(!cycle->Alive()) {
                alpha -= 2 * (currentTime - cycle->DeathTime());
                if(alpha <= 0) continue;
            }
            glColor4f(cycle->color_.r, cycle->color_.g, cycle->color_.b, alpha);
            eCoord pos = cycle->Position(), dir = cycle->Direction();
            glPushMatrix();
            GLfloat m[16] = {
                                xscale * dir.x, yscale * dir.y, 0, 0,
                                -xscale * dir.y, yscale * dir.x, 0, 0,
                                0, 0, 1, 0,
                                pos.x, pos.y, 0, 1
                            };
            glMultMatrixf(m);
            glBegin(GL_TRIANGLES);
            glVertex2f(.5, 0);
            glVertex2f(-.5, .5);
            glVertex2f(-.5, -.5);
            glEnd();
            glPopMatrix();
        }
    }
}

/*
* rimWalls    draw rim walls? 
* cycleWalls  draw cycle walls? 
* cycles      draw cycles? 
* cycleSize   size of cycles in real-world units 
* border      border around bounds (in grid units) 
* x y w h     the map will always fit in this rectangle (in OpenGL units) 
* rw rh       width and height of the rectangle in real-world units 
* ix iy       position where inside? (range [0, 1], used if there's room left due to nonmatching aspect ratio) 
*/ 
void DrawMap(bool rimWalls, bool cycleWalls, bool cycles,
             double cycleSize, double border,
             double x, double y, double w, double h,
             double rw, double rh, double ix, double iy) {
    if(!rimWalls && !cycleWalls && !cycles) return;
    const eRectangle &bounds = eWallRim::GetBounds();
    double lx = bounds.GetLow().x - border, hx = bounds.GetHigh().x + border;
    double ly = bounds.GetLow().y - border, hy = bounds.GetHigh().y + border;
    double mw = hx - lx, mh = hy - ly;
    double xpos, ypos, xscale, yscale;
    double xrat = (rw * mh) / (rh * mw);
    double yrat = (rh * mw) / (rw * mh);
    if(xrat > yrat) {
        xscale = (w * rh) / (mh * rw);
        yscale = h / mh;
        xpos = x + ix * (w - w * yrat);
        ypos = y;
    } else {
        xscale = w / mw;
        yscale = (h * rw) / (mw * rh);
        xpos = x;
        ypos = y + iy * (h - h * xrat);
    }
    glPushMatrix();
    GLfloat m[16] = {
                        xscale, 0, 0, 0,
                        0, yscale, 0, 0,
                        0, 0, 1, 0,
                        xpos - xscale * lx, ypos - yscale * ly, 0, 1
                    };
    glMultMatrixf(m);
    if(rimWalls)
        DrawRimWalls(se_rimWalls);
    if(cycleWalls) {
        DrawWalls(gridded_sg_netPlayerWalls);
        DrawWalls(sg_netPlayerWalls);
    }
    if(cycles)
        DrawCycles(se_PlayerNetIDs, (cycleSize * w) / (rw * xscale), (cycleSize * h) / (rh * yscale));
    glPopMatrix();
}

static void drawsimplemap() {
    // I haven't checked possible initial matrix state, so init to identity and modelview
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    glDisable(GL_TEXTURE_2D);
    glDisable(GL_LIGHTING);
    if(simplemapmode & 1)
        DrawMap(true, true, true,
                5.5, 10,
                .5, -1, .5, .5 * sr_screenWidth / sr_screenHeight,
                .25 * sr_screenWidth, .25 * sr_screenWidth, 1, 0);
    if(simplemapmode & 2)
        DrawMap(true, true, true,
                5.5, 10,
                -1, -1, 2, 2,
                sr_screenWidth, sr_screenHeight, .5, .5);
}

static rPerFrameTask simplemaptask(&drawsimplemap);
#endif	// HUD_MAP

#endif // DEDICATED
